{$MODE objfpc}
{$PACKRECORDS C}

//from mdbtools.h 2010 Theo Lustenberger

(* MDB Tools - A library for reading MS Access database files
 * Copyright (C) 2000 Brian Bruns
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *)

unit mdbtools;

interface

uses  unixtype, glib, glib2;

const
  C_MDB_DEBUG = 0;
  MDB_PGSIZE = 4096;
  MDB_MAX_OBJ_NAME = 256;
  MDB_MAX_COLS = 256;
  MDB_MAX_IDX_COLS = 10;
  MDB_CATALOG_PG = 18;
  MDB_MEMO_OVERHEAD = 12;
  MDB_BIND_SIZE = 16384;

  MDB_VER_JET3 = 0;
  MDB_VER_JET4 = 1;

  MDB_FORM = 0;
  MDB_TABLE = 1;
  MDB_MACRO = 2;
  MDB_SYSTEM_TABLE = 3;
  MDB_REPORT = 4;
  MDB_QUERY = 5;
  MDB_LINKED_TABLE = 6;
  MDB_MODULE = 7;
  MDB_RELATIONSHIP = 8;
  MDB_UNKNOWN_09 = 9;
  MDB_UNKNOWN_0A = 10;
  MDB_DATABASE_PROPERTY = 1;
  MDB_ANY = -1;

  MDB_BOOL = $01;
  MDB_BYTE = $02;
  MDB_INT = $03;
  MDB_LONGINT = $04;
  MDB_MONEY = $05;
  MDB_FLOAT = $06;
  MDB_DOUBLE = $07;
  MDB_SDATETIME = $08;
  MDB_TEXT = $0A;
  MDB_OLE = $0B;
  MDB_MEMO = $0C;
  MDB_REPID = $0F;
  MDB_NUMERIC = $10;

  MDB_MAX_INDEX_DEPTH = 10;

type

  iconv_t = pointer;

  MdbBackendType = record
    name: ^char;
    needs_length: byte;
    needs_scale: byte;
    needs_quotes: byte;
  end;

  MdbBackend = record
    types_table: ^MdbBackendType;
  end;

  MdbHandle = record
    f: ^MdbFile;
    cur_pg: guint32;
    row_num: guint16;
    cur_pos: dword;
    pg_buf: array[0..(MDB_PGSIZE) - 1] of byte;
    alt_pg_buf: array[0..(MDB_PGSIZE) - 1] of byte;
    num_catalog: dword;
    catalog: PGPtrArray;
    default_backend: ^MdbBackend;
    backend_name: ^char;
    fmt: ^MdbFormatConstants;
    stats: ^MdbStatistics;
    iconv_in: iconv_t;
    iconv_out: iconv_t;
  end;

  MdbCatalogEntry = record
    mdb: ^MdbHandle;
    object_name: array[0..(MDB_MAX_OBJ_NAME + 1) - 1] of char;
    object_type: longint;
    table_pg: dword;
    kkd_pg: dword;
    kkd_rowid: dword;
    num_props: longint;
    props: ^TGArray;
    columns: ^TGArray;
    flags: longint;
  end;

  MdbProperties = record
    name: ^gchar;
    hash: ^TGHashTable;
  end;

  MdbAny = record
    case longint of
      0: (i: longint);
      1: (d: double);
      2: (s: array[0..255] of char);
  end;

  MdbColumn = record
    name: array[0..(MDB_MAX_OBJ_NAME + 1) - 1] of char;
    col_type: longint;
    col_size: longint;
    bind_ptr: pointer;
    len_ptr: ^longint;
    properties: ^TGHashTable;
    num_sargs: dword;
    sargs: ^TGPtrArray;
    idx_sarg_cache: ^TGPtrArray;
    is_fixed: byte;
    query_order: longint;
    col_num: longint;
    cur_value_start: longint;
    cur_value_len: longint;
    cur_blob_pg_row: guint32;
    chunk_size: longint;
    col_prec: longint;
    col_scale: longint;
    props: ^MdbProperties;
    fixed_offset: longint;
    var_col_num: dword;
    row_col_num: longint;
  end;

  mdbsargtree = record
    op: longint;
    col: ^MdbColumn;
    value: MdbAny;
    parent: pointer;
    left: ^MdbSargNode;
    right: ^MdbSargNode;
  end;

  MdbIndexPage = record
    pg: guint32;
    start_pos: longint;
    offset: longint;
    len: longint;
    idx_starts: array[0..1999] of guint16;
    cache_value: array[0..255] of byte;
  end;

  MdbFile = record
    fd: longint;
    writable: gboolean;
    filename: ^char;
    jet_version: guint32;
    db_key: guint32;
    db_passwd: array[0..13] of char;
    default_backend: ^MdbBackend;
    backend_name: ^char;
    stats: ^MdbStatistics;
    map_sz: longint;
    free_map: ^byte;
    refs: longint;
  end;

  MdbStatistics = record
    collect: gboolean;
    pg_reads: dword;
  end;

  MdbFormatConstants = record
    pg_size: ssize_t;
    row_count_offset: guint16;
    tab_num_rows_offset: guint16;
    tab_num_cols_offset: guint16;
    tab_num_idxs_offset: guint16;
    tab_num_ridxs_offset: guint16;
    tab_usage_map_offset: guint16;
    tab_first_dpg_offset: guint16;
    tab_cols_start_offset: guint16;
    tab_ridx_entry_size: guint16;
    col_fixed_offset: guint16;
    col_size_offset: guint16;
    col_num_offset: guint16;
    tab_col_entry_size: guint16;
    tab_free_map_offset: guint16;
    tab_col_offset_var: guint16;
    tab_col_offset_fixed: guint16;
    tab_row_col_num_offset: guint16;
  end;

  MdbSargNode = mdbsargtree;

  MdbSargTreeFunc = function(_para1: MdbSargNode; data: gpointer): longint; cdecl;

  MdbStrategy = (MDB_TABLE_SCAN, MDB_LEAF_SCAN, MDB_INDEX_SCAN);

  MdbFileFlags = (MDB_NOFLAGS := $00, MDB_WRITABLE := $01);

  MdbIndexChain = record
    cur_depth: longint;
    last_leaf_found: guint32;
    clean_up_mode: longint;
    pages: array[0..(MDB_MAX_INDEX_DEPTH) - 1] of MdbIndexPage;
  end;

  MdbTableDef = record
    entry: ^MdbCatalogEntry;
    name: array[0..(MDB_MAX_OBJ_NAME + 1) - 1] of char;
    num_cols: dword;
    columns: ^TGPtrArray;
    num_rows: dword;
    index_start: longint;
    num_real_idxs: dword;
    num_idxs: dword;
    indices: ^TGPtrArray;
    first_data_pg: guint32;
    cur_pg_num: guint32;
    cur_phys_pg: guint32;
    cur_row: dword;
    noskip_del: longint;
    map_base_pg: guint32;
    map_sz: size_t;
    usage_map: ^byte;
    freemap_base_pg: guint32;
    freemap_sz: size_t;
    free_usage_map: ^byte;
    sarg_tree: ^MdbSargNode;
    strategy: MdbStrategy;
    scan_idx: ^MdbIndex;
    mdbidx: ^MdbHandle;
    chain: ^MdbIndexChain;
    props: ^MdbProperties;
    num_var_cols: dword;
    is_temp_table: dword;
    temp_table_pages: ^TGPtrArray;
  end;

  mdbindex = record
    index_num: longint;
    name: array[0..(MDB_MAX_OBJ_NAME + 1) - 1] of char;
    index_type: byte;
    first_pg: guint32;
    num_rows: longint;
    num_keys: dword;
    key_col_num: array[0..(MDB_MAX_IDX_COLS) - 1] of smallint;
    key_col_order: array[0..(MDB_MAX_IDX_COLS) - 1] of byte;
    flags: byte;
    table: ^MdbTableDef;
  end;

  MdbColumnProp = record
    name: array[0..(MDB_MAX_OBJ_NAME + 1) - 1] of char;
  end;

  MdbField = record
    value: pointer;
    siz: longint;
    start: longint;
    is_null: byte;
    is_fixed: byte;
    colnum: longint;
    offset: longint;
  end;

  MdbSarg = record
    op: longint;
    value: MdbAny;
  end;

  { mem.c  }

  PMdbBackend = ^MdbBackend;
  PMdbBackendType = ^MdbBackendType;
  PMdbCatalogEntry = ^MdbCatalogEntry;
  PMdbColumn = ^MdbColumn;
  PMdbField = ^MdbField;
  PMdbHandle = ^MdbHandle;
  PMdbIndex = ^MdbIndex;
  PMdbIndexChain = ^MdbIndexChain;
  PMdbIndexPage = ^MdbIndexPage;
  PMdbProperties = ^MdbProperties;
  PMdbSarg = ^MdbSarg;
  PMdbSargNode = ^MdbSargNode;
  PMdbTableDef = ^MdbTableDef;

  function IS_JET4(mdb:PMdbHandle): Boolean;
  function IS_JET3(mdb:PMdbHandle): Boolean;
  procedure Loadmdbtools(lib: pchar);
  procedure Freemdbtools;


var
  mdb_init: procedure; cdecl;
  mdb_exit: procedure; cdecl;

  { file.c  }
  mdb_read_pg: function(mdb: pMdbHandle; pg: dword): ssize_t; cdecl;
  mdb_read_alt_pg: function(mdb: pMdbHandle; pg: dword): ssize_t; cdecl;
  mdb_get_byte: function(buf: pointer; offset: longint): byte; cdecl;
  mdb_get_int16: function(buf: pointer; offset: longint): longint; cdecl;
  mdb_get_int32: function(buf: pointer; offset: longint): longint; cdecl;
  mdb_get_int32_msb: function(buf: pointer; offset: longint): longint; cdecl;
  mdb_get_single: function(buf: pointer; offset: longint): single; cdecl;
  mdb_get_double: function(buf: pointer; offset: longint): double; cdecl;
  mdb_pg_get_byte: function(mdb: pMdbHandle; offset: longint): byte; cdecl;
  mdb_pg_get_int16: function(mdb: pMdbHandle; offset: longint): longint; cdecl;
  mdb_pg_get_int32: function(mdb: pMdbHandle; offset: longint): longint; cdecl;
  mdb_pg_get_single: function(mdb: pMdbHandle; offset: longint): single; cdecl;
  mdb_pg_get_double: function(mdb: pMdbHandle; offset: longint): double; cdecl;
  mdb_open: function(filename: pchar; flags: MdbFileFlags): PMdbHandle; cdecl;
  mdb_close: procedure(mdb: pMdbHandle); cdecl;
  mdb_clone_handle: function(mdb: pMdbHandle): PMdbHandle; cdecl;
  mdb_swap_pgbuf: procedure(mdb: pMdbHandle); cdecl;

  { catalog.c  }
  mdb_free_catalog: procedure(mdb: pMdbHandle); cdecl;
  mdb_read_catalog: function(mdb: pMdbHandle; obj_type: longint): PGPtrArray; cdecl;
  mdb_dump_catalog: procedure(mdb: pMdbHandle; obj_type: longint); cdecl;
  mdb_get_objtype_string: function(obj_type: longint): Pchar; cdecl;

  { table.c  }
  mdb_alloc_tabledef: function(entry: pMdbCatalogEntry): PMdbTableDef; cdecl;
  mdb_free_tabledef: procedure(table: pMdbTableDef); cdecl;
  mdb_read_table: function(entry: pMdbCatalogEntry): PMdbTableDef; cdecl;
  mdb_read_table_by_name: function(mdb: pMdbHandle; table_name: pgchar; obj_type: longint): PMdbTableDef; cdecl;
  mdb_append_column: procedure(columns: pGPtrArray; in_col: pMdbColumn); cdecl;
  mdb_free_columns: procedure(columns: pGPtrArray); cdecl;
  mdb_read_columns: function(table: pMdbTableDef): PGPtrArray; cdecl;
  mdb_table_dump: procedure(entry: pMdbCatalogEntry); cdecl;
  read_pg_if_8: function(mdb: pMdbHandle; cur_pos: plongint): guint8; cdecl;
  read_pg_if_16: function(mdb: pMdbHandle; cur_pos: plongint): guint16; cdecl;
  read_pg_if_32: function(mdb: pMdbHandle; cur_pos: plongint): guint32; cdecl;
  read_pg_if_n: function(mdb: pMdbHandle; buf: pointer; cur_pos: plongint; len: size_t): pointer; cdecl;
  mdb_is_user_table: function(entry: pMdbCatalogEntry): boolean; cdecl;
  mdb_is_system_table: function(entry: pMdbCatalogEntry): boolean; cdecl;

  { data.c  }
  mdb_bind_column_by_name: function(table: pMdbTableDef; col_name: pgchar; bind_ptr: pointer; len_ptr: plongint): longint; cdecl;
  mdb_data_dump: procedure(table: pMdbTableDef); cdecl;
  mdb_bind_column: procedure(table: pMdbTableDef; col_num: longint; bind_ptr: pointer; len_ptr: plongint); cdecl;
  mdb_rewind_table: function(table: pMdbTableDef): longint; cdecl;
  mdb_fetch_row: function(table: pMdbTableDef): boolean; cdecl;
  mdb_is_fixed_col: function(col: pMdbColumn): boolean; cdecl;
  mdb_col_to_string: function(mdb: pMdbHandle; buf: pointer; start: longint; datatype: longint; size: longint): Pchar; cdecl;
  mdb_find_pg_row: function(mdb: pMdbHandle; pg_row: longint; buf: Ppointer; off: plongint; len: psize_t): longint; cdecl;
  mdb_find_row: function(mdb: pMdbHandle; row: longint; start: plongint; len: psize_t): longint; cdecl;
  mdb_find_end_of_row: function(mdb: pMdbHandle; row: longint): longint; cdecl;
  mdb_col_fixed_size: function(col: pMdbColumn): longint; cdecl;
  mdb_col_disp_size: function(col: pMdbColumn): longint; cdecl;
  mdb_ole_read_next: function(mdb: pMdbHandle; col: pMdbColumn; ole_ptr: pointer): size_t; cdecl;
  mdb_ole_read: function(mdb: pMdbHandle; col: pMdbColumn; ole_ptr: pointer; chunk_size: longint): size_t; cdecl;
  mdb_set_date_fmt: procedure(_para1: Pchar); cdecl;
  mdb_read_row: function(table: pMdbTableDef; row: dword): longint; cdecl;

  { dump.c  }
  buffer_dump: procedure(buf: pointer; start: longint; len: size_t); cdecl;

  { backend.c  }
  mdb_get_coltype_string: function(backend: pMdbBackend; col_type: longint): Pchar; cdecl;
  mdb_coltype_takes_length: function(backend: pMdbBackend; col_type: longint): longint; cdecl;
  mdb_init_backends: procedure; cdecl;
  mdb_register_backend: procedure(backend: pMdbBackendType; backend_name: pchar); cdecl;
  mdb_remove_backends: procedure; cdecl;
  mdb_set_default_backend: function(mdb: pMdbHandle; backend_name: pchar): longint; cdecl;
  mdb_get_relationships: function(mdb: pMdbHandle): Pchar; cdecl;

  { sargs.c  }
  mdb_test_sargs: function(table: pMdbTableDef; fields: pMdbField; num_fields: longint): longint; cdecl;
  mdb_test_sarg: function(mdb: pMdbHandle; col: pMdbColumn; node: pMdbSargNode; field: pMdbField): longint; cdecl;
  mdb_sql_walk_tree: procedure(node: pMdbSargNode; func: MdbSargTreeFunc; data: gpointer); cdecl;
  mdb_find_indexable_sargs: function(node: pMdbSargNode; data: gpointer): longint; cdecl;
  mdb_add_sarg_by_name: function(table: pMdbTableDef; colname: pchar; in_sarg: pMdbSarg): longint; cdecl;
  mdb_test_string: function(node: pMdbSargNode; s: pchar): longint; cdecl;
  mdb_test_int: function(node: pMdbSargNode; i: gint32): longint; cdecl;
  mdb_add_sarg: function(col: pMdbColumn; in_sarg: pMdbSarg): longint; cdecl;

  { index.c  }
  mdb_read_indices: function(table: pMdbTableDef): PGPtrArray; cdecl;
  mdb_index_dump: procedure(table: pMdbTableDef; idx: pMdbIndex); cdecl;
  mdb_index_scan_free: procedure(table: pMdbTableDef); cdecl;
  mdb_index_find_next_on_page: function(mdb: pMdbHandle; ipg: pMdbIndexPage): longint; cdecl;
  mdb_index_find_next: function(mdb: pMdbHandle; idx: pMdbIndex; chain: pMdbIndexChain; pg: pguint32; row: pguint16): longint; cdecl;
  mdb_index_hash_text: procedure(text: pchar; hash: pchar); cdecl;
  mdb_index_scan_init: procedure(mdb: pMdbHandle; table: pMdbTableDef); cdecl;
  mdb_index_find_row: function(mdb: pMdbHandle; idx: pMdbIndex; chain: pMdbIndexChain; pg: guint32; row: guint16): longint; cdecl;
  mdb_index_swap_n: procedure(src: pbyte; sz: longint; dest: pbyte); cdecl;
  mdb_free_indices: procedure(indices: pGPtrArray); cdecl;
  mdb_index_page_reset: procedure(ipg: pMdbIndexPage);
  mdb_index_pack_bitmap: function(mdb: pMdbHandle; ipg: pMdbIndexPage): longint; cdecl;

  { stats.c  }
  mdb_stats_on: procedure(mdb: pMdbHandle); cdecl;
  mdb_stats_off: procedure(mdb: pMdbHandle); cdecl;
  mdb_dump_stats: procedure(mdb: pMdbHandle); cdecl;

  { like.c  }
  mdb_like_cmp: function(s: pchar; r: pchar): longint; cdecl;

  { write.c  }
  mdb_crack_row: function(table: pMdbTableDef; row_start: longint; row_end: longint; fields: pMdbField): longint; cdecl;
  mdb_add_row_to_pg: function(table: pMdbTableDef; row_buffer: pbyte; new_row_size: longint): guint16; cdecl;
  mdb_update_index: function(table: pMdbTableDef; idx: pMdbIndex; num_fields: dword; fields: pMdbField; pgnum: guint32;
    rownum: guint16): longint; cdecl;
  mdb_pack_row: function(table: pMdbTableDef; row_buffer: pbyte; num_fields: dword; fields: pMdbField): longint; cdecl;
  mdb_replace_row: function(table: pMdbTableDef; row: longint; new_row: pointer; new_row_size: longint): longint; cdecl;
  mdb_pg_get_freespace: function(mdb: pMdbHandle): longint; cdecl;
  mdb_update_row: function(table: pMdbTableDef): longint; cdecl;
  mdb_new_data_pg: function(entry: pMdbCatalogEntry): pointer; cdecl;

  mdb_insert_row: function(table: pMdbTableDef; num_fields:integer; fields: pMdbField): integer; cdecl;
  _mdb_put_int32: procedure(buf:Pointer; offset:guint32; value: guint32); cdecl;

  { map.c  }
  mdb_map_find_next_freepage: function(table: pMdbTableDef; row_size: longint): guint32; cdecl;
  mdb_map_find_next: function(mdb: pMdbHandle; map: pbyte; map_sz: dword; start_pg: guint32): guint32; cdecl;

  { props.c  }
  mdb_read_props_list: function(kkd: pgchar; len: longint): PGPtrArray; cdecl;
  mdb_free_props: procedure(props: pMdbProperties); cdecl;
  mdb_read_props: function(mdb: pMdbHandle; names: pGPtrArray; kkd: pgchar; len: longint): PMdbProperties; cdecl;

  { worktable.c  }
  mdb_create_temp_table: function(mdb: pMdbHandle; name: pchar): PMdbTableDef; cdecl;
  mdb_temp_table_add_col: procedure(table: pMdbTableDef; col: pMdbColumn); cdecl;
  mdb_fill_temp_col: procedure(tcol: pMdbColumn; col_name: pchar; col_size: longint; col_type: longint; is_fixed: longint); cdecl;
  mdb_fill_temp_field: procedure(field: pMdbField; value: pointer; siz: longint; is_fixed: longint; is_null: longint;
    start: longint; column: longint); cdecl;
  mdb_temp_columns_end: procedure(table: pMdbTableDef); cdecl;

  { options.c  }
  mdb_get_option: function(optnum: dword): longint; cdecl;
  mdb_debug_a: procedure(klass: longint; fmt: pchar; args: array of const); cdecl;
  mdb_debug: procedure(klass: longint; fmt: pchar); cdecl;

  { iconv.c  }
  mdb_unicode2ascii: function(mdb: pMdbHandle; src: pchar; slen: size_t; dest: pchar; dlen: size_t): longint; cdecl;
  mdb_ascii2unicode: function(mdb: pMdbHandle; src: pchar; slen: size_t; dest: pchar; dlen: size_t): longint; cdecl;
  mdb_iconv_init: procedure(mdb: pMdbHandle); cdecl;
  mdb_iconv_close: procedure(mdb: pMdbHandle); cdecl;


implementation

uses
  SysUtils, dynlibs;

var
  hlib: tlibhandle;


function IS_JET4(mdb:PMdbHandle): Boolean;
begin
  Result:=mdb^.f^.jet_version=MDB_VER_JET4;
end;

function IS_JET3(mdb:PMdbHandle): Boolean;
begin
  Result:=mdb^.f^.jet_version=MDB_VER_JET3;
end;


procedure Freemdbtools;
begin
  if hlib<>0 then FreeLibrary(hlib);
  mdb_init := nil;
  mdb_exit := nil;
  mdb_read_pg := nil;
  mdb_read_alt_pg := nil;
  mdb_get_byte := nil;
  mdb_get_int16 := nil;
  mdb_get_int32 := nil;
  mdb_get_int32_msb := nil;
  mdb_get_single := nil;
  mdb_get_double := nil;
  mdb_pg_get_byte := nil;
  mdb_pg_get_int16 := nil;
  mdb_pg_get_int32 := nil;
  mdb_pg_get_single := nil;
  mdb_pg_get_double := nil;
  mdb_open := nil;
  mdb_close := nil;
  mdb_clone_handle := nil;
  mdb_swap_pgbuf := nil;
  mdb_free_catalog := nil;
  mdb_read_catalog := nil;
  mdb_dump_catalog := nil;
  mdb_get_objtype_string := nil;
  mdb_alloc_tabledef := nil;
  mdb_free_tabledef := nil;
  mdb_read_table := nil;
  mdb_read_table_by_name := nil;
  mdb_append_column := nil;
  mdb_free_columns := nil;
  mdb_read_columns := nil;
  mdb_table_dump := nil;
  read_pg_if_8 := nil;
  read_pg_if_16 := nil;
  read_pg_if_32 := nil;
  read_pg_if_n := nil;
  mdb_is_user_table := nil;
  mdb_is_system_table := nil;
  mdb_bind_column_by_name := nil;
  mdb_data_dump := nil;
  mdb_bind_column := nil;
  mdb_rewind_table := nil;
  mdb_fetch_row := nil;
  mdb_is_fixed_col := nil;
  mdb_col_to_string := nil;
  mdb_find_pg_row := nil;
  mdb_find_row := nil;
  mdb_find_end_of_row := nil;
  mdb_col_fixed_size := nil;
  mdb_col_disp_size := nil;
  mdb_ole_read_next := nil;
  mdb_ole_read := nil;
  mdb_set_date_fmt := nil;
  mdb_read_row := nil;
  buffer_dump := nil;
  mdb_get_coltype_string := nil;
  mdb_coltype_takes_length := nil;
  mdb_init_backends := nil;
  mdb_register_backend := nil;
  mdb_remove_backends := nil;
  mdb_set_default_backend := nil;
  mdb_get_relationships := nil;
  mdb_test_sargs := nil;
  mdb_test_sarg := nil;
  mdb_sql_walk_tree := nil;
  mdb_find_indexable_sargs := nil;
  mdb_add_sarg_by_name := nil;
  mdb_test_string := nil;
  mdb_test_int := nil;
  mdb_add_sarg := nil;
  mdb_read_indices := nil;
  mdb_index_dump := nil;
  mdb_index_scan_free := nil;
  mdb_index_find_next_on_page := nil;
  mdb_index_find_next := nil;
  mdb_index_hash_text := nil;
  mdb_index_scan_init := nil;
  mdb_index_find_row := nil;
  mdb_index_swap_n := nil;
  mdb_free_indices := nil;
  mdb_index_page_reset := nil;
  mdb_index_pack_bitmap := nil;
  mdb_stats_on := nil;
  mdb_stats_off := nil;
  mdb_dump_stats := nil;
  mdb_like_cmp := nil;
  mdb_crack_row := nil;
  mdb_add_row_to_pg := nil;
  mdb_update_index := nil;
  mdb_pack_row := nil;
  mdb_replace_row := nil;
  mdb_pg_get_freespace := nil;
  mdb_update_row := nil;
  mdb_new_data_pg := nil;
  mdb_map_find_next_freepage := nil;
  mdb_map_find_next := nil;
  mdb_read_props_list := nil;
  mdb_free_props := nil;
  mdb_read_props := nil;
  mdb_create_temp_table := nil;
  mdb_temp_table_add_col := nil;
  mdb_fill_temp_col := nil;
  mdb_fill_temp_field := nil;
  mdb_temp_columns_end := nil;
  mdb_get_option := nil;
  mdb_debug := nil;
  mdb_debug := nil;
  mdb_unicode2ascii := nil;
  mdb_ascii2unicode := nil;
  mdb_iconv_init := nil;
  mdb_iconv_close := nil;
end;


procedure Loadmdbtools(lib: pchar);
begin
  Freemdbtools;
  hlib := LoadLibrary(lib);
  if hlib = 0 then
    raise Exception.Create(format('Could not load library: %s', [lib]));

  pointer(mdb_init) := GetProcAddress(hlib, 'mdb_init');
  pointer(mdb_exit) := GetProcAddress(hlib, 'mdb_exit');
  pointer(mdb_read_pg) := GetProcAddress(hlib, 'mdb_read_pg');
  pointer(mdb_read_alt_pg) := GetProcAddress(hlib, 'mdb_read_alt_pg');
  pointer(mdb_get_byte) := GetProcAddress(hlib, 'mdb_get_byte');
  pointer(mdb_get_int16) := GetProcAddress(hlib, 'mdb_get_int16');
  pointer(mdb_get_int32) := GetProcAddress(hlib, 'mdb_get_int32');
  pointer(mdb_get_int32_msb) := GetProcAddress(hlib, 'mdb_get_int32_msb');
  pointer(mdb_get_single) := GetProcAddress(hlib, 'mdb_get_single');
  pointer(mdb_get_double) := GetProcAddress(hlib, 'mdb_get_double');
  pointer(mdb_pg_get_byte) := GetProcAddress(hlib, 'mdb_pg_get_byte');
  pointer(mdb_pg_get_int16) := GetProcAddress(hlib, 'mdb_pg_get_int16');
  pointer(mdb_pg_get_int32) := GetProcAddress(hlib, 'mdb_pg_get_int32');
  pointer(mdb_pg_get_single) := GetProcAddress(hlib, 'mdb_pg_get_single');
  pointer(mdb_pg_get_double) := GetProcAddress(hlib, 'mdb_pg_get_double');
  pointer(mdb_open) := GetProcAddress(hlib, 'mdb_open');
  pointer(mdb_close) := GetProcAddress(hlib, 'mdb_close');
  pointer(mdb_clone_handle) := GetProcAddress(hlib, 'mdb_clone_handle');
  pointer(mdb_swap_pgbuf) := GetProcAddress(hlib, 'mdb_swap_pgbuf');
  pointer(mdb_free_catalog) := GetProcAddress(hlib, 'mdb_free_catalog');
  pointer(mdb_read_catalog) := GetProcAddress(hlib, 'mdb_read_catalog');
  pointer(mdb_dump_catalog) := GetProcAddress(hlib, 'mdb_dump_catalog');
  pointer(mdb_get_objtype_string) := GetProcAddress(hlib, 'mdb_get_objtype_string');
  pointer(mdb_alloc_tabledef) := GetProcAddress(hlib, 'mdb_alloc_tabledef');
  pointer(mdb_free_tabledef) := GetProcAddress(hlib, 'mdb_free_tabledef');
  pointer(mdb_read_table) := GetProcAddress(hlib, 'mdb_read_table');
  pointer(mdb_read_table_by_name) := GetProcAddress(hlib, 'mdb_read_table_by_name');
  pointer(mdb_append_column) := GetProcAddress(hlib, 'mdb_append_column');
  pointer(mdb_free_columns) := GetProcAddress(hlib, 'mdb_free_columns');
  pointer(mdb_read_columns) := GetProcAddress(hlib, 'mdb_read_columns');
  pointer(mdb_table_dump) := GetProcAddress(hlib, 'mdb_table_dump');
  pointer(read_pg_if_8) := GetProcAddress(hlib, 'read_pg_if_8');
  pointer(read_pg_if_16) := GetProcAddress(hlib, 'read_pg_if_16');
  pointer(read_pg_if_32) := GetProcAddress(hlib, 'read_pg_if_32');
  pointer(read_pg_if_n) := GetProcAddress(hlib, 'read_pg_if_n');
  pointer(mdb_is_user_table) := GetProcAddress(hlib, 'mdb_is_user_table');
  pointer(mdb_is_system_table) := GetProcAddress(hlib, 'mdb_is_system_table');
  pointer(mdb_bind_column_by_name) := GetProcAddress(hlib, 'mdb_bind_column_by_name');
  pointer(mdb_data_dump) := GetProcAddress(hlib, 'mdb_data_dump');
  pointer(mdb_bind_column) := GetProcAddress(hlib, 'mdb_bind_column');
  pointer(mdb_rewind_table) := GetProcAddress(hlib, 'mdb_rewind_table');
  pointer(mdb_fetch_row) := GetProcAddress(hlib, 'mdb_fetch_row');
  pointer(mdb_is_fixed_col) := GetProcAddress(hlib, 'mdb_is_fixed_col');
  pointer(mdb_col_to_string) := GetProcAddress(hlib, 'mdb_col_to_string');
  pointer(mdb_find_pg_row) := GetProcAddress(hlib, 'mdb_find_pg_row');
  pointer(mdb_find_row) := GetProcAddress(hlib, 'mdb_find_row');
  pointer(mdb_find_end_of_row) := GetProcAddress(hlib, 'mdb_find_end_of_row');
  pointer(mdb_col_fixed_size) := GetProcAddress(hlib, 'mdb_col_fixed_size');
  pointer(mdb_col_disp_size) := GetProcAddress(hlib, 'mdb_col_disp_size');
  pointer(mdb_ole_read_next) := GetProcAddress(hlib, 'mdb_ole_read_next');
  pointer(mdb_ole_read) := GetProcAddress(hlib, 'mdb_ole_read');
  pointer(mdb_set_date_fmt) := GetProcAddress(hlib, 'mdb_set_date_fmt');
  pointer(mdb_read_row) := GetProcAddress(hlib, 'mdb_read_row');
  pointer(buffer_dump) := GetProcAddress(hlib, 'buffer_dump');
  pointer(mdb_get_coltype_string) := GetProcAddress(hlib, 'mdb_get_coltype_string');
  pointer(mdb_coltype_takes_length) := GetProcAddress(hlib, 'mdb_coltype_takes_length');
  pointer(mdb_init_backends) := GetProcAddress(hlib, 'mdb_init_backends');
  pointer(mdb_register_backend) := GetProcAddress(hlib, 'mdb_register_backend');
  pointer(mdb_remove_backends) := GetProcAddress(hlib, 'mdb_remove_backends');
  pointer(mdb_set_default_backend) := GetProcAddress(hlib, 'mdb_set_default_backend');
  pointer(mdb_get_relationships) := GetProcAddress(hlib, 'mdb_get_relationships');
  pointer(mdb_test_sargs) := GetProcAddress(hlib, 'mdb_test_sargs');
  pointer(mdb_test_sarg) := GetProcAddress(hlib, 'mdb_test_sarg');
  pointer(mdb_sql_walk_tree) := GetProcAddress(hlib, 'mdb_sql_walk_tree');
  pointer(mdb_find_indexable_sargs) := GetProcAddress(hlib, 'mdb_find_indexable_sargs');
  pointer(mdb_add_sarg_by_name) := GetProcAddress(hlib, 'mdb_add_sarg_by_name');
  pointer(mdb_test_string) := GetProcAddress(hlib, 'mdb_test_string');
  pointer(mdb_test_int) := GetProcAddress(hlib, 'mdb_test_int');
  pointer(mdb_add_sarg) := GetProcAddress(hlib, 'mdb_add_sarg');
  pointer(mdb_read_indices) := GetProcAddress(hlib, 'mdb_read_indices');
  pointer(mdb_index_dump) := GetProcAddress(hlib, 'mdb_index_dump');
  pointer(mdb_index_scan_free) := GetProcAddress(hlib, 'mdb_index_scan_free');
  pointer(mdb_index_find_next_on_page) := GetProcAddress(hlib, 'mdb_index_find_next_on_page');
  pointer(mdb_index_find_next) := GetProcAddress(hlib, 'mdb_index_find_next');
  pointer(mdb_index_hash_text) := GetProcAddress(hlib, 'mdb_index_hash_text');
  pointer(mdb_index_scan_init) := GetProcAddress(hlib, 'mdb_index_scan_init');
  pointer(mdb_index_find_row) := GetProcAddress(hlib, 'mdb_index_find_row');
  pointer(mdb_index_swap_n) := GetProcAddress(hlib, 'mdb_index_swap_n');
  pointer(mdb_free_indices) := GetProcAddress(hlib, 'mdb_free_indices');
  pointer(mdb_index_page_reset) := GetProcAddress(hlib, 'mdb_index_page_reset');
  pointer(mdb_index_pack_bitmap) := GetProcAddress(hlib, 'mdb_index_pack_bitmap');
  pointer(mdb_stats_on) := GetProcAddress(hlib, 'mdb_stats_on');
  pointer(mdb_stats_off) := GetProcAddress(hlib, 'mdb_stats_off');
  pointer(mdb_dump_stats) := GetProcAddress(hlib, 'mdb_dump_stats');
  pointer(mdb_like_cmp) := GetProcAddress(hlib, 'mdb_like_cmp');
  pointer(mdb_crack_row) := GetProcAddress(hlib, 'mdb_crack_row');
  pointer(mdb_add_row_to_pg) := GetProcAddress(hlib, 'mdb_add_row_to_pg');
  pointer(mdb_update_index) := GetProcAddress(hlib, 'mdb_update_index');
  pointer(mdb_pack_row) := GetProcAddress(hlib, 'mdb_pack_row');
  pointer(mdb_replace_row) := GetProcAddress(hlib, 'mdb_replace_row');
  pointer(mdb_pg_get_freespace) := GetProcAddress(hlib, 'mdb_pg_get_freespace');
  pointer(mdb_update_row) := GetProcAddress(hlib, 'mdb_update_row');

  pointer(mdb_insert_row) := GetProcAddress(hlib, 'mdb_insert_row');
  pointer(_mdb_put_int32) := GetProcAddress(hlib, '_mdb_put_int32');

  pointer(mdb_new_data_pg) := GetProcAddress(hlib, 'mdb_new_data_pg');
  pointer(mdb_map_find_next_freepage) := GetProcAddress(hlib, 'mdb_map_find_next_freepage');
  pointer(mdb_map_find_next) := GetProcAddress(hlib, 'mdb_map_find_next');
  pointer(mdb_read_props_list) := GetProcAddress(hlib, 'mdb_read_props_list');
  pointer(mdb_free_props) := GetProcAddress(hlib, 'mdb_free_props');
  pointer(mdb_read_props) := GetProcAddress(hlib, 'mdb_read_props');
  pointer(mdb_create_temp_table) := GetProcAddress(hlib, 'mdb_create_temp_table');
  pointer(mdb_temp_table_add_col) := GetProcAddress(hlib, 'mdb_temp_table_add_col');
  pointer(mdb_fill_temp_col) := GetProcAddress(hlib, 'mdb_fill_temp_col');
  pointer(mdb_fill_temp_field) := GetProcAddress(hlib, 'mdb_fill_temp_field');
  pointer(mdb_temp_columns_end) := GetProcAddress(hlib, 'mdb_temp_columns_end');
  pointer(mdb_get_option) := GetProcAddress(hlib, 'mdb_get_option');
  pointer(mdb_debug_a) := GetProcAddress(hlib, 'mdb_debug');
  pointer(mdb_debug) := GetProcAddress(hlib, 'mdb_debug');
  pointer(mdb_unicode2ascii) := GetProcAddress(hlib, 'mdb_unicode2ascii');
  pointer(mdb_ascii2unicode) := GetProcAddress(hlib, 'mdb_ascii2unicode');
  pointer(mdb_iconv_init) := GetProcAddress(hlib, 'mdb_iconv_init');
  pointer(mdb_iconv_close) := GetProcAddress(hlib, 'mdb_iconv_close');
end;


initialization
  hlib:=0;
//  Loadmdbtools('/home/theo/svn/mdbtools/src/libmdb/.libs/libmdb.so');
finalization
  Freemdbtools;

end.
